/* File: WINDOW.C */

#include "work.h"
#include <string.h>
#include "ERROR.h"
#include "WINDOW.h"

XFontStruct *WINDOW::QueryFont (void)
{
   XFontStruct *font_info = XQueryFont (display, GetFont ());

   if (font_info == NULL) {
      ERROR ("XQueryFont");
      return NULL;
   }
   return font_info;
}

void WINDOW::SendEvent (XEvent *event)
{
   if (XSendEvent (display, parent ->WindowID (), False, NoEventMask, event)
      == 0) {
      ERROR ("XSendEvent");
   }
}

void WINDOW::SetCursor (Cursor cursor)
{
   XDefineCursor (display, window, cursor);
}

void WINDOW::RestoreCursor (void)
{
   XUndefineCursor (display, window);
}

void WINDOW::Lower (void)
{
   XLowerWindow (display, window);
}

void WINDOW::Raise (void)
{
   XRaiseWindow (display, window);
}

Window WINDOW::WindowID (void)
{
   return window;
}

void WINDOW::SetName (char *name)
{
   XStoreName (display, window, name);
}

char * WINDOW::GetName (void)
{
   char *name;

   if (XFetchName (display, window, &name) == 0) {
      ERROR ("XFetchName");
   }
   return name;
}

void WINDOW::GetAttributes (XWindowAttributes *xwa)
{
   if (XGetWindowAttributes (display, window, xwa) == 0) 
      ERROR ("XGetWindowAttributes");
}

void WINDOW::SetNormalHints (XSizeHints *hint)
{
   hint ->flags = PPosition | PSize;
   XSetWMNormalHints (display, window, hint);
}

void WINDOW::MoveResize (int x, int y, unsigned int width, unsigned int height)
{
   XMoveResizeWindow (display, window, x, y, width, height);
}

void WINDOW::Resize (unsigned int width, unsigned int height)
{
   XResizeWindow (display, window, width, height);
}

void WINDOW::Move (int x, int y)
{
   XMoveWindow (display, window, x, y);
}

void WINDOW::GetNormalHints (XSizeHints *hint)
{
   if (XGetWMNormalHints (display, window, hint, (long *) hint) == 0) {
      ERROR ("XGetWMNormalHints");
   }
}

void WINDOW::SetInput (long mask)
{
   XSelectInput (display, window, mask);
}

void WINDOW::SetFont (Font fid)
{
   XSetFont (display, gc, fid);
}

Font WINDOW::GetFont (void)
{
   XGCValues t;

   XGetGCValues (display, gc, GCFont, &t);
   return t .font;
}

void WINDOW::SetBackground (unsigned long col)
{
   XSetBackground (display, gc, col);
}

unsigned long WINDOW::GetBackground (void)
{
   XGCValues t;

   XGetGCValues (display, gc, GCBackground, &t);
   return t .background;
}

void WINDOW::SetForeground (unsigned long col)
{
   XSetForeground (display, gc, col);
}

unsigned long WINDOW::GetForeground (void)
{
   XGCValues t;

   XGetGCValues (display, gc, GCForeground, &t);
   return t .foreground;
}

void WINDOW::DrawString (int x, int y, char *string)
{
   XDrawImageString (display, window, gc, x, y, string, strlen (string));
}

void WINDOW::DrawLine (int x1, int y1, int x2, int y2)
{
   XDrawLine (display, window, gc, x1, y1, x2, y2);
}

void WINDOW::DrawRectangle (int x, int y, unsigned int width,
   unsigned int height)
{
   XDrawRectangle (display, window, gc, x, y, width, height);
}

void WINDOW::FillRectangle (int x, int y, unsigned int width,
   unsigned int height)
{
   XFillRectangle (display, window, gc, x, y, width, height);
}

void WINDOW::Clear (void)
{
   XClearWindow (display, window);
}

WINDOW::~WINDOW (void)
{
   XFreeGC (display, gc);
   XSync (display, True);
//   printf ("destroy window %ul\n", window);
   XDestroyWindow (display, window);
}

void WINDOW::SetVisible (Bool vsbl)
{
   visible = vsbl;

   Window       *children;
   unsigned int n_children;

   GetChildren (&children, &n_children);

   for (int i = 0; i < n_children; i++) {
      struct {
         unsigned char *tc;
      } *prop;

      GetThisProperty (display, children [i], (unsigned char **) &prop);
      ((WINDOW *) (prop ->tc)) ->SetVisible (vsbl);
      XFree ((char *) prop);
   }
   if (n_children > 0) XFree ((char *) children);
}

void WINDOW::Show (Bool vsbl)
{
   if (!visible) return;
   if (vsbl) {
//      printf ("mapping window %ul\n", window);
      XMapWindow (display, window);
   }

   Window       *children;
   unsigned int n_children;

   GetChildren (&children, &n_children);

   for (int i = 0; i < n_children; i++) {
      struct {
         unsigned char *tc;
      } *prop;

      GetThisProperty (display, children [i], (unsigned char **) &prop);
      ((WINDOW *) (prop ->tc)) ->Show (vsbl);
      XFree ((char *) prop);
   }
   if (n_children > 0) XFree ((char *) children);
   if (!vsbl) {
//      printf ("unmapping window %ul\n", window);
      XUnmapWindow (display, window);
   }
}

WINDOW::WINDOW (Display *dsp, WINDOW *prnt, char *ClassName, char *Name, int x,
   int y, unsigned int width, unsigned int height, unsigned int border_width,
   unsigned long border, unsigned long background, Bool vsbl, Font fid,
   unsigned long foreground)
{
   display    = dsp;
   parent     = prnt;

   Window parentWID = prnt == NULL ? DefaultRootWindow (display) :
      prnt ->WindowID ();

   window     = XCreateSimpleWindow (dsp, parentWID,
      x, y, width, height, border_width, border, background);

   if (window == 0) {
      ERROR ("WINDOW");
   }
   strcpy (class_name, ClassName);
   gc         = XCreateGC (display, window, 0, 0);
   if (gc == 0) {
      ERROR ("WINDOW");
   }
   visible    = vsbl;
   if (Name != NULL) SetName (Name);

   union {
      unsigned long tl;
      void          *tc;
   } tmp;

   tmp .tc = this;
   SetThisProperty (display, window, (unsigned char *) &(tmp .tl));
   SetBackground (background);
   SetForeground (foreground);
   SetFont (fid);

   XSizeHints hint;

   hint .x      = x;
   hint .y      = y;
   hint .width  = width;
   hint .height = height;
   SetNormalHints (&hint);
}

void WINDOW::GetChildren (Window **children, unsigned int *n_children)
{
   Window root, parent;

   if (XQueryTree (display, window, &root, &parent, children, n_children) 
      == 0) {
      ERROR ("XQueryTree");
   }
}

int Dispatch (XEvent *event)
{
   Window  win;
   Display *display;

   switch (event ->type) {
   case KeyPress:
   case KeyRelease:
      win     = event ->xkey .window;
      display = event ->xkey .display;
      break;
   case ButtonPress:
   case ButtonRelease:
      win     = event ->xbutton .window;
      display = event ->xbutton .display;
      break;
   case EnterNotify:
   case LeaveNotify:
      win     = event ->xcrossing .window;
      display = event ->xcrossing .display;
      break;
   case Expose:
      win     = event ->xexpose .window;
      display = event ->xexpose .display;
      break;
   case NoExpose:
      win     = event ->xnoexpose .drawable;
      display = event ->xnoexpose .display;
      break;
   case ClientMessage:
      win     = event ->xclient .window;
      display = event ->xclient .display;
      break;
   default:
      {
         ERROR ("Dispatch: unknown message");
      }
      break;
   }

   struct {
      unsigned char *tc;
   } *prop;

   GetThisProperty (display, win, (unsigned char **) &prop);

   int ret = ((WINDOW *) (prop ->tc)) ->Go (event);

   XFree ((char *) prop);
   return ret;
}

