/* File : exkor.C */

#include <stdio.h>
#include <malloc.h>
#include <values.h>
#include <stdlib.h>
#include <math.h>
#include "const.h"
#include "res.h"
#include "task.h"
#include "ERROR.h"
#include "OUTPUT.h"

#define CON           1.28
#define LTONE         7.
#define UTZERO        18.66
#define V1            0.398942280444
#define V2            0.399903438504
#define V3            5.75885480458
#define V4            29.8213557808
#define V5            2.62433121679
#define V6            48.6959930692
#define V7            5.92885724438
#define V11           0.398942280385
#define V12           3.8052e-8
#define V13           1.00000615302
#define V14           3.98063794e-4
#define V15           1.98615381364
#define V16           0.151679116635
#define V17           5.29330324926
#define V18           4.8385912808
#define V19           15.1508972451
#define V20           0.742380924027
#define V21           30.789933034
#define V22           3.99019417011

#define RMAXS sqrt (r1mach (2))
#define RMAX  log10 (r1mach (2))
#define K     ((int) RMAX + 1)
#define CUND  -2.0 * K
#define K1    (K / 3)
#define CP5   pow (10., K1 - 1.)
#define CP6   (10. * CP5)
#define CM6   (1.0 / CP6)
#define CM5   (1.0 / CP5)
#define DM17  pow (10.0, 2. - K)
#define CM17  DM17

static double wiener_process_parameter (const double *x, const double *y,
   int n);
static double probability (const double *x, const double *y, double ym,
   double e, double a, const double *z, int n);
static double alnorm (double x);
static void cor (double x1, double x2, double y3, double y2, double de,
   double yh, double *x, double *y);
static double calculate_yh (double ym, double ymax);
static double extrapolate (double a, double b);
static void unimodality (double *z, const double *x, const double *y, int j,
   double eb2);
static int extrk0 (double lb, double ub, double *ym,
   int max_n_of_fun_eval, int n_of_init_points, double accuracy_of_y,
   double accuracy_of_x, int coord, double *x, int n, 
   double (*f) (const double *x, int n),
   double *x1, double *y3, double *x2, double *y2, int *n_of_fun_eval);
static void exkor0 (double *x, const double *a, const double *b, int n,
   double *y, double (*f) (const double *x, int n), int max_l, 
   int init_points);

static int    n_of_points;
static double *fm;
static double **xn;

/* estimation of Wiener process parameter */

static double wiener_process_parameter (const double *x, const double *y,
   int n)
{
   double d = 0.;
   int    j;

   for (j = 1; j < n; j++) {
      double td = y [j] - y [j - 1];

      d += td * td / (x [j] - x [j - 1]);
   }
   return (sqrt (d / (n - 1)));
}

/*
   probability of evaluating ym with accuracy e
   model of function is Wiener process with parameter a
   function is not in the interval of unimodality
   z [i] > 0. stands for the interval of unimodality
*/

static double probability (const double *x, const double *y, double ym,
   double e, double a, const double *z, int n)
{
   double p4 = 1.;
   double p  = - 800. / a * a;

   n--;
   while (n-- > 0) {
      if (z [n] < 0.) {
         double p1 = p;

         p1 *= (y [n + 1] - ym + e) * (y [n] - ym + e);
         p1 /= x[n + 1] - x [n];
         if (p1 > CUND) {
            p4 *= 1. - exp (p1);
         }
      }
   }
   return (p4);
}

/* the Gausian integral */

static double alnorm (double x)
{
   int    up = FALSE;
   double ret;

   if (x < 0.) {
      x  = - x;
      up = TRUE;
   }
   if (x <= LTONE || up == TRUE && x <= UTZERO) {
      double y = 0.5 * x * x;

      if (x <= CON) {
         ret = y + V7;
         ret = y + V5 + V6 / ret;
         ret = y + V3 - V4 / ret;
         ret = 0.5 - x * (V1 - V2 * y / ret);
      } else {
         ret = x + V22;
         ret = x + V20 + V21 / ret;
         ret = x + V18 - V19 / ret;
         ret = x - V16 + V17 / ret;
         ret = x + V14 + V15 / ret;
         ret = x - V12 + V13 / ret;
         ret = V11 * exp (-y) / ret;
      }
   } else {
      ret = 0;
   }
   if (up == FALSE) {
      ret = 1. - ret;
   }
   return (ret);
}

static void cor (double x1, double x2, double y3, double y2, double de,
   double yh, double *x, double *y)
{
   double am, d, x12, xx1, xx2;

   x12 = x2 - x1;
   *x  = 0.5 * (x1 + x2 + x12 * (y3 - y2) / (y3 + y2 - 2 * yh));
   xx2 = x2 - (*x);
   xx1 = *x - x1;
   am  = (y3 * xx2 +y2 * xx1) / x12;
   d   = sqrt (xx2 * xx1 / x12) * de;
   *y  = - alnorm ((yh - am) / d);
}

static double calculate_yh (double ym, double ymax)
{
   return (ym - 0.5 * (ymax - ym));
}

static double extrapolate (double a, double b)
{
   return (0.5 * (3 * a - b));
}

static void unimodality (double *z, const double *x, const double *y, int j,
   double eb2)
{
   if (z [j + 1] > 0. &&
       y [j] > y [j + 1] &&
       (x [j + 1] - x [j]) < eb2) {
      z [j] = 1.;
   }
}

static int extrk0 (double lb, double ub, double *ym,
   int max_n_of_fun_eval, int n_of_init_points, double accuracy_of_y,
   double accuracy_of_x, int coord, double *x, int n, 
   double (*f) (const double *x, int n),
   double *x1, double *y3, double *x2, double *y2, int *n_of_fun_eval)
{
   double a, ymax, p1, p4, xa, ya, interval, eb2, temp, da, p2, yh;
   double store_x, store_y, xm;
   int    j, i, n4, ns, n5, n7, n8, n1, loc_optimization;
   int    ret_code = 0;

/*
   ns - number of really evaluated points in the set x1
   n1 - number of function evaluations
*/
   store_x   = xm = x [coord];
   store_y  = *ym;
   ns       = n1 = 0;
   interval = ub - lb;
   eb2      = interval * 0.04;
   if (interval < CM5 || interval > CP5) {
      ERROR (
"WARNING: bounds are too small or too large", False);
      goto end;
   }
   da  = interval / (n_of_init_points - 1);
   xa  = lb;
   for (j = 3; j < n_of_init_points + 3; j++) {
      x [coord] = x1 [j] = xa;
      ns++;
      fm [n_of_points] = ya = y3 [j] = (*f) (x, n);
      copy_array (xn [n_of_points], x, n);
      n_of_points++;
      n1++;
      if (fabs (ya) > RMAXS) {
         ERROR (
"ERROR: function value too large", False);
         ret_code = -1;
         goto end;
      }
      if (ya < *ym) {
         *ym    = ya;
         xm     = xa;
      }
      if (met_out (n_of_points) == EXIT_MET) {
         ret_code = -1;
         goto end;
      }
      xa += da;
   }
/*
   a is parameter of Wiener process
   parameter of model double (*f) (const double *x, int n)
*/
   a  = wiener_process_parameter (x1 + 3, y3 + 3, n_of_init_points) * 20.;
   if (a < CM6 || a > CP6) {
      ERROR (
"WARNING: variation of function is too small or too large", False);
      goto end;
   }
   array_maximum (y3 + 3, n_of_init_points, &j);
   ymax          = y3 [j + 3];
   yh            = calculate_yh (*ym, ymax);
/*
   a is evaluated. n5 is the number of local optima
   computed with accuracy accuracy_of_y, accuracy_of_x
*/
   n5            = 0;
   accuracy_of_y = MIN (accuracy_of_y, 0.1 * a * sqrt (eb2));
   xa            = store_x;
   p2            = interval * 0.001;
/*
   optimal point from previous searches is included in the point set if it is
   rather distant from other points in the set.
   if p1 is near some of points x1 [j] (distance from the point is not greater
   than interval / 1000) then nothing is done;
   else point is included to the set x1 (the set is widened by 1)
*/
   for (j = 3; j < ns + 3; j++) {
      if (xa > x1 [j + 1]) continue;
      if (xa <= x1 [j] + p2 || xa >= x1 [j + 1] - p2) break;
      for (i = ns + 2; i > j; i--) {
         x1 [i + 1] = x1 [i];
         y3 [i + 1] = y3 [i];
      }
      ns++;
      x1 [j + 1] = xa;
      y3 [j + 1] = store_y;
      if (store_y < *ym) {
         *ym = store_y;
         xm  = xa;
         yh  = calculate_yh (*ym, ymax);
      }
      break;
   }
   double_array_fill (y2, max_n_of_fun_eval + 6, 0.);
/* Bayesian step - find interval of biggest improvement  */
   for (;;) {
      for (j = 3; j < ns + 2; j++)
         if (y2 [j] <= 0.)
            cor (x1 [j], x1 [j + 1],
                 y3 [j], y3 [j + 1],
                 a     , yh        ,
                 x2 + j, y2 + j);
      do {
         array_minimum (y2 + 3, ns - 1, &j);
         p1  = y2 [j + 3];
         j  += 3;
/*
   widen set x1 by 1 point and include it in the place, where y2 achieves
   minimum, the new point is x2 [j]
*/
         for (i = ns + 2; i > j; i--) {
            x1 [i + 1] = x1 [i];
            y3 [i + 1] = y3 [i];
            x2 [i]     = x2 [i - 1];
            y2 [i]     = y2 [i - 1];
         }
         x1 [j + 1] = x [coord] = xa = x2 [j];
         ns++;
         fm [n_of_points] = y3 [j + 1] = ya = (*f) (x, n);
         copy_array (xn [n_of_points], x, n);
         n_of_points++;
         n1++;
         if (fabs (ya) > RMAXS) {
            ERROR (
"ERROR: function value too large", False);
            ret_code = -1;
            goto end;
         }
         if (ya < *ym) {
            *ym    = ya;
            xm     = xa;
            n4     = 2;
         } else {
            n4     = 1;
         }
         if (met_out (n_of_points) == EXIT_MET) {
            ret_code = -1;
            goto end;
         }
/* if number of evaluations of f more than max_n_of_fun_eval - 1 go to end  */
         if (n1 >= max_n_of_fun_eval) goto end;
         ymax = MAX (ymax, ya);
/*
   p4 is probability of evaluating
   xm, ym with accuracy accuracy_of_y, accuracy_of_x
*/
         p4 = probability (x1 + 3, y3 + 3, *ym, accuracy_of_y, a, y2 + 3, ns);
         yh = calculate_yh (*ym, ymax);
         if (fabs (*ym - ya) < r1mach (1) * 1.0e+3) {
            n4 = 2;
            a  = wiener_process_parameter (x1 + 3, y3 + 3, ns) * 20.;
         } else {
/* if p4 >= 0.95 go to end  */
            if (p4 >= 0.95) goto end;
/* if n5 then a is updated in each fifth step */
            if (ns % 5 == 0 && n5 == 0) {
               n4 = 2;
               a  = wiener_process_parameter (x1 + 3, y3 + 3, ns) * 20.;
            }
         }
/*
   preparation of bounds of x1, y3 for testing -
   is interval of local optimum found
*/
         x1 [0]      = x1 [1]      = x1 [2]      =
            extrapolate (x1 [3], x1 [4]);
         x1 [ns + 3] = x1 [ns + 4] = x1 [ns + 5] =
            extrapolate (x1 [ns + 2], x1 [ns + 1]);
         y3 [0]      = y3 [1]      = y3 [2]      = y3 [4];
         y3 [ns + 3] = y3 [ns + 4] = y3 [ns + 5] = y3 [ns + 1];
/*  test of interval of local minimum found  */
         temp             = interval * 0.2;
         loc_optimization = FALSE;
         for (i = MAX (0, j - 5); i < MIN (ns, j + 2); i++) {
            if (y3 [i    ] <  y3 [i + 1]) continue;
            if (y3 [i + 1] <  y3 [i + 2]) continue;
            if (y3 [i + 2] <= y3 [i + 3]) continue;
            if (y3 [i + 3] >= y3 [i + 4]) continue;
            if (y3 [i + 4] >  y3 [i + 5]) continue;
            if (y3 [i + 5] >  y3 [i + 6]) continue;
            if ((x1 [i + 6] - x1 [i]) <= temp) {
               loc_optimization = TRUE;
               break;
            }
         }
/*   begining of local optimization   */
         if (loc_optimization == TRUE) {
            double xl1, xl2, xl3, yl1, yl2, yl3, p3;

            double_array_fill (y2 + i, 6, 1.);
            i   += 3;
            n5++;
            p3   = x1 [0];
            xl1  = x1 [i - 1];
            xl2  = x1 [i];
            xl3  = x1 [i + 1];
            yl1  = y3 [i - 1];
            yl2  = y3 [i];
            yl3  = y3 [i + 1];
            n8   = 0;
            for (;;) {
               double ee1, ee2, ee3, ee4, ee;

               ee1 = yl1 - yl2;
               ee2 = xl1 - xl2;
               if (fabs (ee1) <= DM17 ||
                   fabs (ee2) <= DM17) break;
               ee3 = yl1 - yl3;
               ee4 = xl1 - xl3;
               ee  = 2 * (ee3 / ee1 - ee4 / ee2);
               if (fabs (ee)  <= DM17) break;
               xa  = ((xl1 + xl2) * ee3 / ee1 - (xl1 + xl3) * ee4 / ee2) / ee;
               ee3 = xl2 - xa;
               ee2 = xl1 - xa;
               if (fabs (ee2) < DM17) {
                  p1 = yl1;
               } else {
                  ee3 = 1 - (ee3 / ee2) * (ee3 / ee2);
                  if (fabs (ee3) < DM17) {
                     p1 = (ee1 > fabs (ee3) * 1.0e+6) ?
                          (yl1 - 1.0e+5): (yl1 - ee1 / ee3);
                  } else {
                     p1 = yl1 - ee1 / ee3;
                  }
               }
               x [coord]        = xa;
               fm [n_of_points] = ya = (*f) (x, n);
               copy_array (xn [n_of_points], x, n);
               n_of_points++;
               n1++;
               if (fabs (ya) > RMAXS) {
                  ERROR (
"ERROR: function value too large", False);
                  ret_code = -1;
                  goto end;
               }
               if (ya < *ym) {
                  *ym    = ya;
                  xm     = xa;
                  yh     = calculate_yh (*ym, ymax);
               }
               if (met_out (n_of_points) == EXIT_MET) {
                  ret_code = -1;
                  goto end;
               }
               if (n1 >= max_n_of_fun_eval) goto end;
               p3 = fabs (xa - p3);
               n7 = (xa > xl2) ? 2 : 1;
               p1 = fabs (p1 - ya);
               if ((p1 <= accuracy_of_y &&
                    p3 <= accuracy_of_x) ||
                   p1 <= CM17            ||
                   p3 <= CM17) break;
               if (ya < yl2) {
                  if (n7 == 2) {
                     xl1 = xl2;
                     yl1 = yl2;
                     xl2 = xa;
                     yl2 = ya;
                     if (i == ns + 2 && n8 == 0) {
                        xl3 = extrapolate (xl2, xl1);
                        yl3 = yl1;
                     } else {
                        n8 = 1;
                     }
                  } else {
                     xl3 = xl2;
                     yl3 = yl2;
                     xl2 = xa;
                     yl2 = ya;
                     if (i == 3 && n8 == 0) {
                        xl1 = extrapolate (xl2, xl3);
                        yl1 = yl3;
                     } else {
                        n8 = 1;
                     }
                  }
               } else {
                  if (n7 == 2) {
                     xl3 = xa;
                     yl3 = ya;
                     if (i == 3 && n8 == 0) {
                        xl1 = extrapolate (xl2, xl3);
                        yl1 = yl3;
                     } else {
                        n8 = 1;
                     }
                  } else {
                     xl1 = xa;
                     yl1 = ya;
                     if (i == ns + 2 && n8 == 0) {
                        xl3 = extrapolate (xl2, xl1);
                        yl3 = yl1;
                     } else {
                        n8 = 1;
                     }
                  }
               }
               p3 = xa;
            }
            x1 [i] = xa;
            y3 [i] = ya;
            n4     = 2;
         }
         if (n4 == 2) break;
         for (i = j; i < j + 2; i++) {
            double xj1, xj2, yj1, yj2;

            xj1 = x1 [i];
            xj2 = x1 [i + 1];
            yj1 = y3 [i];
            yj2 = y3 [i + 1];
            if (xj2 - xj1 < eb2 && y2 [i - 1] > 0. && yj1 < yj2) {
               y2 [i] = 1.;
            } else {
               cor (xj1, xj2, yj1, yj2, a, yh, x2 + i, y2 + i);
            }
         }
         unimodality (y2, x1, y3, j + 1, eb2);
         unimodality (y2, x1, y3, j    , eb2);
      } while (n4 == 1);
   }
end:
   *n_of_fun_eval = n1;
   x [coord]      = xm;
   return ret_code;
}

/*---------------------------------------------------------------------------*/

static double *x1 = NULL;
static double *y3, *y2, *x2;

#define I100 100 /* max number of function evaluations in line search */

void exkor (double *x, const double *a, const double *b, int n, double *y,
   double (*f) (const double *x, int n), int max_l, int init_points, 
   double *pfm, double **pxn)
{
   fm = pfm;
   xn = pxn;
   if ((x1 = (double *)
      malloc (4 * (I100 + 6) * sizeof (double))) == NULL) {
      ERROR ("out of memory", False);
      goto ret;
   }
   x2 = x1 + I100 + 6;
   y2 = x2 + I100 + 6;
   y3 = y2 + I100 + 6;
   exkor0 (x, a, b, n, y, f, max_l, init_points);
ret:
   if (x1 != NULL) {
      free ((char *) x1);
      x1 = NULL;
   }
}

static void exkor0 (double *x, const double *a, const double *b, int n,
   double *y, double (*f) (const double *x, int n), int max_l,
   int init_points)
{
   int coord, n_of_fun_eval, max_n_of_fun_eval;

   n_of_points      = 0;
   fm [n_of_points] = *y = (*f) (x, n);
   copy_array (xn [n_of_points], x, n);
   n_of_points++;
   if (met_out (n_of_points)== EXIT_MET) goto end;
   max_l--;
   max_n_of_fun_eval = MIN (I100, max_l);
   for (;;) {
      for (coord = 0; coord < n; coord++) {
         if (extrk0 (a [coord], b [coord], y, max_n_of_fun_eval,
            init_points, 1.0e-6, 1.0e-6, coord, x, n, f, x1, y3, y2, x2,
            &n_of_fun_eval) == -1) goto end;
         max_l -= n_of_fun_eval;
         if (max_l <= 0) goto end;
         max_n_of_fun_eval = MIN (I100, max_l);
      }
   }
end:;
}


